/***********************************************************************
MPEG-4 Audio RM Module
Lossless coding of 1-bit oversampled audio - DST (Direct Stream Transfer)

This software was originally developed by:

* Aad Rijnberg 
  Philips Digital Systems Laboratories Eindhoven 
  <aad.rijnberg@philips.com>

* Fons Bruekers
  Philips Research Laboratories Eindhoven
  <fons.bruekers@philips.com>
   
* Eric Knapen
  Philips Digital Systems Laboratories Eindhoven
  <h.w.m.knapen@philips.com> 

And edited by:

* Richard Theelen
  Philips Digital Systems Laboratories Eindhoven
  <r.h.m.theelen@philips.com>

in the course of development of the MPEG-4 Audio standard ISO-14496-1, 2 and 3.
This software module is an implementation of a part of one or more MPEG-4 Audio
tools as specified by the MPEG-4 Audio standard. ISO/IEC gives users of the
MPEG-4 Audio standards free licence to this software module or modifications
thereof for use in hardware or software products claiming conformance to the
MPEG-4 Audio standards. Those intending to use this software module in hardware
or software products are advised that this use may infringe existing patents.
The original developers of this software of this module and their company,
the subsequent editors and their companies, and ISO/EIC have no liability for
use of this software module or modifications thereof in an implementation.
Copyright is not released for non MPEG-4 Audio conforming products. The
original developer retains full right to use this code for his/her own purpose,
assign or donate the code to a third party and to inhibit third party from
using the code for non MPEG-4 Audio conforming products. This copyright notice
must be included in all copies of derivative works.

Copyright  2004.

Source file: FIR.c (Prediction filter)

Required libraries: <none>

Authors:
RT:  Richard Theelen, PDSL-labs Eindhoven <r.h.m.theelen@philips.com>

Changes:
08-Mar-2004 RT  Initial version

************************************************************************/


/*============================================================================*/
/*       INCLUDES                                                             */
/*============================================================================*/


#include "math.h"

#define IMIN MIN
#define H_IABS(a,b) abs(b)
#define IFLIP(a,b) (((a) == 0) ? (b) : (-(b)))
#define UBYTESEL(a, b) ( ((a) >> ((b)*8)) & 0xFF )

#include "FIR.h"
#include "types.h"

/*============================================================================*/
/*       STATIC VARIABLES                                                     */
/*============================================================================*/

static unsigned char InitStream[] =
{
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
  0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1
};


/*============================================================================*/
/*       GLOBAL FUNCTION IMPLEMENTATIONS                                      */
/*============================================================================*/

/*
 * NAME            : DST_EFirBitPredFilter
 * 
 * PURPOSE         : Predict the next bit by filtering, determine indices for p-tables
 *                  
 * INPUT           : see definition
 *
 * OUTPUT          : see definition
 *
 * INPUT/OUTPUT    : see definition
 *
 * DATA REFERENCES : InitStream[]
 *
 * RETURN          : DST_NOERROR (OK)
 * 
 */

ENCODING_STATUS  DST_EFirBitPredFilter(/* in */
                                       unsigned char  BitStream[MAXCH][MAXCHBITS],
                                       int            ChannelFilter[MAXCH],
                                       int            ICoef[MAXCH][MAXPREDORDER],
                                       int            NrOfChannelBits,
                                       int            NrOfChannels,
                                       int            OptPredOrder[MAXCH],
                                       /* out */
                                       unsigned char  BitResidual[MAXCH][MAXCHBITS],
                                       unsigned char  Z[MAXCH][MAXCHBITS] )
{                          
  int ChNr;
  int FilterNr;
  int PredVal;
  int State;
  int BitNr;
  int BitIndex;


  for(ChNr=0; ChNr<NrOfChannels; ChNr++)
  {
    FilterNr=ChannelFilter[ChNr];

    /* First (OptPredOrder) bits; BitNr indicates the bit to predict */
    for (BitNr=0; BitNr<OptPredOrder[FilterNr]; BitNr++)
    {
      PredVal=0;

      /* Use preceding bits */
      for (State=0, BitIndex=BitNr-1; 
           State<BitNr; 
           State++, BitIndex--)
      {
          if (BitStream[ChNr][BitIndex])
          {
              PredVal +=ICoef[FilterNr][State];
          }
          else
          {
              PredVal -=ICoef[FilterNr][State];
          }
      }

      /* Use initial values of the filter */
      for (State=BitNr, BitIndex=0;
           State<OptPredOrder[FilterNr];
           State++, BitIndex++)
      {
          if (InitStream[BitIndex])
          {
              PredVal +=ICoef[FilterNr][State];
          }
          else
          {
              PredVal -=ICoef[FilterNr][State];
          }
      }

      /* Determine the residual signal and Z */
      if (PredVal>=0)
      {
          /* Prediction: "1" */
          BitResidual[ChNr][BitNr] = (unsigned char)((int)BitStream[ChNr][BitNr]^1);
          Z[ChNr][BitNr] = (unsigned char)MIN( PredVal>>ACQSTEP, MAXPTIND );
      }
      else
      {
          /* Prediction: "0" */
          BitResidual[ChNr][BitNr] = (unsigned char)((int)BitStream[ChNr][BitNr]^0);
          Z[ChNr][BitNr] = (unsigned char) MIN( (-PredVal)>>ACQSTEP, MAXPTIND );
      }
    }

    /* Rest of bits OptPredOrder ... NrOfChannelBits-1;
       BitNr indicates the bit to predict */
    for (BitNr=OptPredOrder[FilterNr]; 
         BitNr<NrOfChannelBits; 
         BitNr++)
    {
      PredVal=0;

      /* Use preceding bits */
      for (State=0, BitIndex=BitNr-1;
           State<OptPredOrder[FilterNr]; 
           State++, BitIndex-- )
      {
          if (BitStream[ChNr][BitIndex])
          {
              PredVal +=ICoef[FilterNr][State];
          }
          else
          {
              PredVal -=ICoef[FilterNr][State];
          }
      }

      /* Determine the residual signal and Z */
      if (PredVal>=0)
      {
          /* Prediction: "1" */
          BitResidual[ChNr][BitNr] = (unsigned char)((int)BitStream[ChNr][BitNr]^1);
          Z[ChNr][BitNr] = (unsigned char) MIN( PredVal>>ACQSTEP, MAXPTIND );
      }
      else
      {
          /* Prediction: "0" */
          BitResidual[ChNr][BitNr] = (unsigned char)((int)BitStream[ChNr][BitNr]^0);
          Z[ChNr][BitNr] = (unsigned char) MIN( (-PredVal)>>ACQSTEP, MAXPTIND );
      }
    }
  }

  return (DST_NOERROR);
}

/* end of file */
